package de.letsbuildacompiler.compiler;

import de.letsbuildacompiler.compiler.exceptions.UndeclaredVariableException;
import de.letsbuildacompiler.compiler.exceptions.UndefinedFunctionException;
import de.letsbuildacompiler.compiler.exceptions.VariableAlreadyDefinedException;
import de.letsbuildacompiler.parser.DemoBaseVisitor;
import de.letsbuildacompiler.parser.DemoParser;
import de.letsbuildacompiler.parser.DemoParser.MainStatementContext;
import de.letsbuildacompiler.parser.DemoParser.NumberContext;
import de.letsbuildacompiler.parser.DemoParser.PlusContext;
import org.antlr.v4.runtime.Token;
import org.antlr.v4.runtime.tree.ParseTree;

import java.util.HashMap;
import java.util.Map;

public class JasminVisitor extends DemoBaseVisitor<String> {

    // Symbol table
    private Map<String, Integer> variables = new HashMap<>();
    private JvmStack jvmStack = new JvmStack();
    private final FunctionList definedFunctions;
    private int branchCounter = 0;
    private int compareCount = 0;
    private int andCounter = 0;
    private int orCounter = 0;

    JasminVisitor(FunctionList definedFunctions) {
        if (definedFunctions == null) {
            throw new NullPointerException("definedFunctions");
        }
        this.definedFunctions = definedFunctions;
    }

    @Override
    public String visitNumber(NumberContext ctx) {
        jvmStack.push(DataType.INT);
        return "ldc " + ctx.number.getText();
    }

    @Override
    public String visitString(DemoParser.StringContext ctx) {
        jvmStack.push(DataType.STRING);
        return "ldc " + ctx.txt.getText();
    }

    private String visitBinaryOperation(String name, DemoParser.ExpressionContext ctx) {
        String instructions = visitChildren(ctx) + "\n" + name;
        jvmStack.pop();
        jvmStack.pop();
        jvmStack.push(DataType.INT);
        return instructions;
    }

    @Override
    public String visitPlus(PlusContext ctx) {
        return visitBinaryOperation("iadd", ctx);
    }

    @Override
    public String visitMinus(DemoParser.MinusContext ctx) {
        return visitBinaryOperation("isub", ctx);
    }

    @Override
    public String visitDiv(DemoParser.DivContext ctx) {
        return visitBinaryOperation("idiv", ctx);
    }

    @Override
    public String visitMult(DemoParser.MultContext ctx) {
        return visitBinaryOperation("imul", ctx);
    }

    @Override
    public String visitRelational(DemoParser.RelationalContext ctx) {
        int compareNum = compareCount++;
        String jumpInstruction;

        switch (ctx.operator.getText()) {
            case "<":
                jumpInstruction = "if_icmplt";
                break;
            case "<=":
                jumpInstruction = "if_icmple";
                break;
            case ">":
                jumpInstruction = "if_icmpgt";
                break;
            case ">=":
                jumpInstruction = "if_icmpge";
                break;
            default:
                throw new IllegalArgumentException("Unknown operator : <" + ctx.operator.getText() + ">");
        }

        final String instructions = visitChildren(ctx) + "\n" +
                jumpInstruction + " onTrue" + compareNum + "\n" +
                "ldc 0\n" +
                "goto onFalse" + compareNum + "\n" +
                "onTrue" + compareNum + ":\n" +
                "ldc 1\n" +
                "onFalse" + compareNum + ":";
        jvmStack.pop();
        jvmStack.pop();
        jvmStack.push(DataType.INT);
        return instructions;
    }

    @Override
    public String visitAnd(DemoParser.AndContext ctx) {
        String left = visit(ctx.left);
        String right = visit(ctx.right);
        int andNum = andCounter++;
        jvmStack.pop();
        jvmStack.pop();
        jvmStack.push(DataType.INT);

        return left + "\n" +
                "ifeq onAndFalse" + andNum + "\n" +
                right + "\n" +
                "ifeq onAndFalse" + andNum + "\n" +
                "ldc 1\n" +
                "goto andEnd" + andNum + "\n" +
                "onAndFalse" + andNum + ":\n" +
                "ldc 0\n" +
                "andEnd" + andNum + ":";
    }

    @Override
    public String visitOr(DemoParser.OrContext ctx) {
        String left = visit(ctx.left);
        String right = visit(ctx.right);
        int orNum = orCounter++;
        jvmStack.pop();
        jvmStack.pop();
        jvmStack.push(DataType.INT);

        return left + "\n" +
                "ifne onOrTrue" + orNum + "\n" +
                right + "\n" +
                "ifne onOrTrue" + orNum + "\n" +
                "ldc 0\n" +
                "goto orEnd" + orNum + "\n" +
                "onOrTrue" + orNum + ":\n" +
                "ldc 1\n" +
                "orEnd" + orNum + ":";
    }

    @Override
    public String visitPrintln(DemoParser.PrintlnContext ctx) {
        String arg = visit(ctx.argument);
        DataType type = jvmStack.pop();
        return "getstatic java/lang/System/out Ljava/io/PrintStream;\n" +
                arg + "\n" +
                "invokevirtual java/io/PrintStream/println(" + type.getJvmType() + ")V\n";
    }

    @Override
    public String visitPrint(DemoParser.PrintContext ctx) {
        String arg = visit(ctx.argument);
        DataType type = jvmStack.pop();
        return "getstatic java/lang/System/out Ljava/io/PrintStream;\n" +
                arg + "\n" +
                "invokevirtual java/io/PrintStream/print(" + type.getJvmType() + ")V\n";
    }

    @Override
    public String visitVariable(DemoParser.VariableContext ctx) {
        jvmStack.push(DataType.INT);
        return "iload " + requireVariableIndex(ctx.varName);
    }

    private int requireVariableIndex(Token varName) {
        Integer varIndex = variables.get(varName.getText());
        if (varIndex == null) {
            throw new UndeclaredVariableException(varName);
        }
        return varIndex;
    }

    @Override
    public String visitVarDeclaration(DemoParser.VarDeclarationContext ctx) {
        String key = ctx.varName.getText();
        if (variables.containsKey(key)) {
            throw new VariableAlreadyDefinedException(ctx.varName);
        }
        variables.put(key, variables.size());
        return "";
    }

    @Override
    public String visitBranch(DemoParser.BranchContext ctx) {
        String condition = visit(ctx.condition);
        jvmStack.pop();
        String onTrue = visit(ctx.onTrue);
        String onFalse = visit(ctx.onFalse);
        int branchNum = branchCounter++;

        return condition + "\n" +
                "ifne ifTrue" + branchNum + "\n" +
                onFalse + "\n" +
                "goto endIf" + branchNum + "\n" +
                "ifTrue" + branchNum + ":\n" +
                onTrue + "\n" +
                "endIf" + branchNum + ":\n";
    }

    @Override
    public String visitAssignment(DemoParser.AssignmentContext ctx) {
        final String instructions = visit(ctx.expr) + "\n"
                + "istore " + requireVariableIndex(ctx.varName) + "\n";
        jvmStack.pop();
        return instructions;
    }

    @Override
    public String visitFunctionDefinition(DemoParser.FunctionDefinitionContext ctx) {
        Map<String, Integer> oldVars = this.variables;
        JvmStack oldJvmStack = jvmStack;
        jvmStack = new JvmStack();
        variables = new HashMap<>();
        visit(ctx.params);
        String statements = visit(ctx.statements);
        String instructions = ".method public static " + ctx.funcName.getText() + "(";
        int numberOfParams = ctx.params.declarations.size();
        instructions += stringRepeat("I", numberOfParams);
        instructions += ")I\n" +
                ".limit locals 100\n" +
                ".limit stack 100\n" +
                (statements == null ? "" : statements + "\n") +
                visit(ctx.retValue) + "\n" +
                "ireturn\n" +
                ".end method\n";
        jvmStack.pop();
        variables = oldVars;
        jvmStack = oldJvmStack;
        return instructions;
    }

    private String stringRepeat(String string, int numberOfParams) {
        StringBuilder result = new StringBuilder();
        for (int i = 0; i < numberOfParams; i++) {
            result.append(string);
        }
        return result.toString();
    }

    @Override
    public String visitFunctionCall(DemoParser.FunctionCallContext ctx) {
        if (!definedFunctions.contains(ctx.funcName.getText(), ctx.args.expressions.size() )) {
            throw new UndefinedFunctionException(ctx.funcName);
        }

        String instructions = "";
        String args = visit(ctx.args);
        if (args != null) {
            instructions += args + "\n";
        }
        instructions += "invokestatic HelloWorld/" + ctx.funcName.getText() + "(";
        int numberOfParams = ctx.args.expressions.size();
        instructions += stringRepeat("I", numberOfParams);
        instructions +=")I\n";
        for (int i = 0; i < numberOfParams; i++) {
            jvmStack.pop();
        }
        jvmStack.push(DataType.INT);
        return instructions;
    }

    @Override
    public String visitProgram(DemoParser.ProgramContext ctx) {
        StringBuilder mainCode = new StringBuilder();
        StringBuilder functions = new StringBuilder();

        for (ParseTree child : ctx.children) {
            final String instructions = visit(child);
            if (child instanceof MainStatementContext) {
                mainCode.append(instructions);
            } else {
                functions.append(instructions);
            }
        }
        final String instructions = ".method public static main([Ljava/lang/String;)V\n" +
                ".limit stack 100\n" +
                ".limit locals 100\n" +
                mainCode + "\n" +
                "return\n" +
                ".end method\n";

        return functions.append(instructions).toString();
    }

    @Override
    protected String aggregateResult(String aggregate, String nextResult) {
        if (aggregate == null) {
            return nextResult;
        }
        if (nextResult == null) {
            return aggregate;
        }
        return aggregate + "\n" + nextResult;
    }
}
