package de.letsbuildacompiler.compiler;

import de.letsbuildacompiler.compiler.exceptions.FunctionAlreadyDefinedException;
import de.letsbuildacompiler.compiler.exceptions.UndeclaredVariableException;
import de.letsbuildacompiler.compiler.exceptions.UndefinedFunctionException;
import de.letsbuildacompiler.compiler.exceptions.VariableAlreadyDefinedException;
import jasmin.ClassFile;
import org.antlr.v4.runtime.CharStreams;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Objects;
import java.util.Scanner;

import static org.testng.AssertJUnit.assertEquals;

public class CompilerTest {

    private Path tempDir;

    @BeforeMethod
    public void createTempDir() throws IOException {
        tempDir = Files.createTempDirectory("compilerTest");
    }

    @AfterMethod
    public void deleteTempDir() {
        deleteRecursive(tempDir.toFile());
    }

    private void deleteRecursive(File file) {
        if (file.isDirectory()) {
            for (File f : Objects.requireNonNull(file.listFiles())) {
                deleteRecursive(f);
            }
        }

        if (!file.delete()) {
            throw new Error("Could not delete file <" + file + ">");
        }
    }

    @Test(dataProvider = "provider")
    @SuppressWarnings("unused")
    public void runningCode_outputsExpectedText(String description, String code, String expectedText) throws Exception {
        String actual = compileAndRun(code);
        assertEquals(expectedText, actual);
    }

    @DataProvider(name = "provider")
    public Object[][] provide_code_expectedText() {
        return new Object[][] {
                {"plus", "println(1+2);", "3" + System.lineSeparator()},
                {"chained plus", "println(1+4+42);", "47" + System.lineSeparator()},
                {"multiple statements", "println(1); println(2);", "1" + System.lineSeparator() + "2" + System.lineSeparator()},
                {"minus", "println(3-2);", "1" + System.lineSeparator()},
                {"times", "println(3*3);", "9" + System.lineSeparator()},
                {"division", "println(6/3);", "2" + System.lineSeparator()},
                {"truncated division", "println(7/2);", "3" + System.lineSeparator()},
                {"operator precedence times and divide", "println(8/2*4);", "16" + System.lineSeparator()},
                {"operator precedence times and plus", "println(2+3*3);", "11" + System.lineSeparator()},
                {"operator precedence times and minus", "println(9-2*3);", "3" + System.lineSeparator()},
                {"operator precedence minus and plus", "println(8-2+5);", "11" + System.lineSeparator()},
                {"int variable", "int foo; foo = 42; println(foo);", "42" + System.lineSeparator()},
                {"add variable and constant parameters",
                        "int foo; foo = 42; println(foo+2);", "44" + System.lineSeparator()},
                {"add two variables parameters",
                        "int foo; int bar; foo = 42; bar = 43; println(foo+bar);", "85" + System.lineSeparator()},

                example("functions/simple_function", "4" + System.lineSeparator()),
                example("functions/scopes", "4" + System.lineSeparator() + "42" + System.lineSeparator()),
                example("functions/int_parameters", "13" + System.lineSeparator()),
                example("branch/if_int_false", "42" + System.lineSeparator()),
                example("branch/if_int_true", "81" + System.lineSeparator()),

                {"lower than true", "println(1 < 2);", "1" + System.lineSeparator()},
                {"lower than false", "println(2 < 2);", "0" + System.lineSeparator()},
                {"lower or equal true", "println(1 <= 1);", "1" + System.lineSeparator()},
                {"lower or equal false", "println(3 <= 2);", "0" + System.lineSeparator()},

                {"greater than true", "println(3 > 2);", "1" + System.lineSeparator()},
                {"greater than false", "println(3 > 3);", "0" + System.lineSeparator()},
                {"greater or equal true", "println(3 >= 3);", "1" + System.lineSeparator()},
                {"greater or equal false", "println(2 >= 3);", "0" + System.lineSeparator()},

                {"and true", "println(1 && 1);", "1" + System.lineSeparator()},
                {"and left false", "println(0 && 1);", "0" + System.lineSeparator()},
                {"and right false", "println(1 && 0);", "0" + System.lineSeparator()},
                example("operators/and-skip-right", "0" + System.lineSeparator() + "0" + System.lineSeparator()),

                {"or false", "println(0 || 0);", "0" + System.lineSeparator()},
                {"or left true", "println(1 || 0);", "1" + System.lineSeparator()},
                {"or right true", "println(0 || 1);", "1" + System.lineSeparator()},
                example("operators/or-skip-right", "1" + System.lineSeparator() + "1" + System.lineSeparator()),

                {"print", "print(42);", "42"},
                {"print string litteral", "print(\"Hello world\");", "Hello world"},
        };
    }

    @Test(expectedExceptions = UndeclaredVariableException.class,
    expectedExceptionsMessageRegExp = "1:8 undeclared variable <x>")
    public void compilingCode_throwsUndeclaredVariableException_ifReadingUndefinedVariable() throws Exception {
        compileAndRun("println(x);");
        // evaluation performed by expected exception.
    }

    @Test(expectedExceptions = UndeclaredVariableException.class,
            expectedExceptionsMessageRegExp = "1:0 undeclared variable <x>")
    public void compilingCode_throwsUndeclaredVariableException_ifWritingUndefinedVariable() throws Exception {
        compileAndRun("x = 5;");
        // evaluation performed by expected exception.
    }

    @Test(expectedExceptions = VariableAlreadyDefinedException.class,
            expectedExceptionsMessageRegExp = "2:4 variable <x> already defined")
    public void compilingCode_throwsAlreadyDefinedException_whenDefiningAlreadyDefinedVariable() throws Exception {
        compileAndRun("int x;" + System.lineSeparator() +
                "int x;");
        // evaluation performed by expected exception.
    }

    @Test(expectedExceptions = UndefinedFunctionException.class,
            expectedExceptionsMessageRegExp = "1:8 call to undefined function : <undefinedFunction>")
    public void compilingCode_throwsUndefinedFunctionException_whenCallingUndefinedFunction() throws Exception {
        compileAndRun("println(undefinedFunction());");
        // evaluation performed by expected exception.
    }

    @Test(expectedExceptions = FunctionAlreadyDefinedException.class,
            expectedExceptionsMessageRegExp = "2:4 function already defined : <x>")
    public void compilingCode_throwsAlreadyDefinedFunctionException_whenDefiningFunctionTwice() throws Exception {
        compileAndRun("int x() { return 42; }\n" +
                "int x() { return 45; }");
        // evaluation performed by expected exception.
    }

    private String compileAndRun(String code) throws Exception {
        code = Main.compile(CharStreams.fromString(code));
        ClassFile classFile = new ClassFile();
        classFile.readJasmin(new StringReader(code), "", false);
        Path outputPath = tempDir.resolve(classFile.getClassName() + ".class");
        try (OutputStream outputStream = Files.newOutputStream(outputPath)) {
            classFile.write(outputStream);
        }
        return runJavaClass(tempDir, classFile.getClassName());
    }

    private String runJavaClass(Path dir, String className) throws IOException {
        Process process = Runtime.getRuntime().exec(new String[]{"java", "-cp", dir.toString(), className});
        try (InputStream in = process.getInputStream()) {
            return new Scanner(in).useDelimiter("\\A").next();
        }
    }

    private String[] example(String name, String expectedResult) throws IllegalArgumentException {
        try (InputStream in = getClass().getResourceAsStream("/examples/" + name)) {
            if (in == null) {
                throw new IllegalArgumentException("No such example " + name);
            }
            String code = new Scanner(in, "UTF-8").useDelimiter("\\A").next();

            return new String[] {name, code, expectedResult};
        } catch (IOException e) {
            throw new IllegalArgumentException("No such example " + name);
        }
    }
}